# =============================================================================
#               ____ _   _ ____  _       _ _ _           _         _ _
#  _ __  _   _ / ___| | | |  _ \| |     | (_) |__   __ _| |__   __| | |
# | '_ \| | | | |  _| |_| | | | | |     | | | '_ \ / _` | '_ \ / _` | |
# | |_) | |_| | |_| |  _  | |_| | |___ _| | | |_) | (_| | | | | (_| | |
# | .__/ \__, |\____|_| |_|____/|_____(_)_|_|_.__/ \__, |_| |_|\__,_|_|
# |_|    |___/                                     |___/
# =============================================================================
#  Authors:
#    Tristan Gingold
#    Patrick Lehmann
#
# Package module:   Python binding and low-level API for shared library 'libghdl'.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================

from ctypes import c_int32

from pyTooling.Decorators import export

from pyGHDL.libghdl import libghdl
from pyGHDL.libghdl._types import (
    NameId,
    Iir_Library_Declaration,
    Iir_Design_Unit,
    Iir_Design_File,
    LocationType,
)
from pyGHDL.libghdl._decorator import BindToLibGHDL

__all__ = ["Library_Location", "Work_Library"]

Library_Location: LocationType = c_int32.in_dll(libghdl, "libraries__library_location")
"""
A location for library declarations (such as library WORK).

Use the property ``.value`` to access the variable's value.
"""

Work_Library: Iir_Library_Declaration = c_int32.in_dll(libghdl, "libraries__work_library")
"""
Library declaration for the work library.

.. note:: The identifier of the work_library is ``work_library_name``, which may be different from 'WORK'.

Use the property ``.value`` to access the variable's value.
"""

Work_Library_Name: NameId = c_int32.in_dll(libghdl, "libraries__work_library_name")
"""
Name of the 'WORK' library, usually 'work'.

Use the property ``.value`` to access the variable's value.
"""


@export
@BindToLibGHDL("libraries__get_libraries_chain")
def Get_Libraries_Chain() -> Iir_Library_Declaration:
    """
    Get the chain of libraries. Can be used only to read (it mustn't be modified).

    :return: undocumented
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("libraries__add_design_unit_into_library")
def Add_Design_Unit_Into_Library(Unit: Iir_Design_Unit, Keep_Obsolete: bool) -> None:
    """
    Add or replace an design unit in the work library. DECL must not have a chain
    (because it may be modified).

    If the design_file of UNIT is not already in the library, a new one is created.

    Units are always appended to the design_file. Therefore, the order is kept.

    :param Unit:          undocumented
    :param Keep_Obsolete: If :obj:`Keep_Obsolete` is True, obsoleted units are
                          kept in the library.

                          This is used when a whole design file has to be added
                          in the library and then processed (without that feature,
                          redefined units would disappear).
    """


@export
@BindToLibGHDL("libraries__purge_design_file")
def Purge_Design_File(Design_File: Iir_Design_File) -> None:
    """
    Remove the same file as :obj:`Design_File` from work library and all of its units.

    :param Design_File: undocumented
    """


@export
@BindToLibGHDL("libraries__find_entity_for_component")
def Find_Entity_For_Component(Name: NameId) -> Iir_Design_Unit:
    """
    Find an entity whose name is :obj:`Name` in any library. |br|
    If there is no such entity, return :attr:`~pyGHDL.libghdl.vhdl.nodes.Null_Iir`. |br|
    If there are several entities, return :attr:`~pyGHDL.libghdl.vhdl.nodes.Null_Iir`;

    :param Name: Entity name to search for.
    :return:     undocumented
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("libraries__get_library_no_create")
def Get_Library_No_Create(Ident: NameId) -> Iir_Library_Declaration:
    """
    Get the library named :obj:`Ident`.

    :param Ident: Library to look for.
    :return:      Return :attr:`~pyGHDL.libghdl.vhdl.nodes.Null_Iir` if it doesn't exist.
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("libraries__get_library")
def Get_Library(Ident: NameId, Loc: LocationType, Force: bool) -> Iir_Library_Declaration:
    """
    Get the library named :obj:`Ident`.

    :param Ident: Library to look for.
    :param Loc: Location in case of errors.
    :return:      Return :attr:`~pyGHDL.libghdl.vhdl.nodes.Null_Iir` if it doesn't exist.
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("libraries__find_primary_unit")
def Find_Primary_Unit(Library: Iir_Library_Declaration, Name: NameId) -> Iir_Design_Unit:
    """
    Just return the design_unit for :obj:`Name`, or ``NULL`` if not found.

    :param Library: Library to look in.
    :param Name:    Primary unit to search for.
    :return:        undocumented
    """
    return 0  # pragma: no cover


@export
@BindToLibGHDL("libraries__load_work_library")
def Load_Work_Library(Empty: bool) -> None:
    """
    Load the work library whose name is ``Work_Library_Name``.

    :param Empty: If set, just create it.
    """



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Package module:   DOM: Concurrent statements.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================
from typing import Iterable

from pyTooling.Decorators import export

from pyVHDLModel.Base import ExpressionUnion, WaveformElement as VHDLModel_WaveformElement
from pyVHDLModel.Symbol import Symbol
from pyVHDLModel.Association import (
    AssociationItem,
    GenericAssociationItem as VHDLModel_GenericAssociationItem,
    PortAssociationItem as VHDLModel_PortAssociationItem,
    ParameterAssociationItem as VHDLModel_ParameterAssociationItem,
)
from pyVHDLModel.Sequential import SequentialStatement
from pyVHDLModel.Concurrent import (
    ComponentInstantiation as VHDLModel_ComponentInstantiation,
    EntityInstantiation as VHDLModel_EntityInstantiation,
    ConfigurationInstantiation as VHDLModel_ConfigurationInstantiation,
    ProcessStatement as VHDLModel_ProcessStatement,
    ConcurrentProcedureCall as VHDLModel_ConcurrentProcedureCall,
    ConcurrentBlockStatement as VHDLModel_ConcurrentBlockStatement,
    IfGenerateBranch as VHDLModel_IfGenerateBranch,
    ElsifGenerateBranch as VHDLModel_ElsifGenerateBranch,
    ElseGenerateBranch as VHDLModel_ElseGenerateBranch,
    IfGenerateStatement as VHDLModel_IfGenerateStatement,
    ConcurrentChoice,
    ConcurrentCase,
    CaseGenerateStatement as VHDLModel_CaseGenerateStatement,
    ForGenerateStatement as VHDLModel_ForGenerateStatement,
    ConcurrentSimpleSignalAssignment as VHDLModel_ConcurrentSimpleSignalAssignment,
    ConcurrentAssertStatement as VHDLModel_ConcurrentAssertStatement,
    ConcurrentStatement,
    GenerateCase as VHDLModel_GenerateCase,
    OthersGenerateCase as VHDLModel_OthersGenerateCase,
    IndexedGenerateChoice as VHDLModel_IndexedGenerateChoice,
    RangedGenerateChoice as VHDLModel_RangedGenerateChoice,
)

from pyGHDL.libghdl import Iir, utils
from pyGHDL.libghdl.vhdl import nodes
from pyGHDL.dom import DOMMixin, DOMException, Position
from pyGHDL.dom.Range import Range
from pyGHDL.dom.Symbol import (
    ArchitectureSymbol,
    EntityInstantiationSymbol,
    ComponentInstantiationSymbol,
    ConfigurationInstantiationSymbol,
)


@export
class GenericAssociationItem(VHDLModel_GenericAssociationItem, DOMMixin):
    def __init__(self, associationNode: Iir, actual: ExpressionUnion, formal: Symbol = None) -> None:
        super().__init__(actual, formal)
        DOMMixin.__init__(self, associationNode)


@export
class PortAssociationItem(VHDLModel_PortAssociationItem, DOMMixin):
    def __init__(self, associationNode: Iir, actual: ExpressionUnion, formal: Symbol = None) -> None:
        super().__init__(actual, formal)
        DOMMixin.__init__(self, associationNode)


@export
class ParameterAssociationItem(VHDLModel_ParameterAssociationItem, DOMMixin):
    def __init__(self, associationNode: Iir, actual: ExpressionUnion, formal: Symbol = None) -> None:
        super().__init__(actual, formal)
        DOMMixin.__init__(self, associationNode)


@export
class ComponentInstantiation(VHDLModel_ComponentInstantiation, DOMMixin):
    def __init__(
        self,
        instantiationNode: Iir,
        label: str,
        componentSymbol: ComponentInstantiationSymbol,
        genericAssociations: Iterable[AssociationItem] = None,
        portAssociations: Iterable[AssociationItem] = None,
    ) -> None:
        super().__init__(label, componentSymbol, genericAssociations, portAssociations)
        DOMMixin.__init__(self, instantiationNode)

    @classmethod
    def parse(cls, instantiationNode: Iir, instantiatedUnit: Iir, label: str) -> "ComponentInstantiation":
        from pyGHDL.dom._Translate import GetName, GetGenericMapAspect, GetPortMapAspect

        componentSymbol = ComponentInstantiationSymbol(instantiatedUnit, GetName(instantiatedUnit))
        genericAssociations = GetGenericMapAspect(nodes.Get_Generic_Map_Aspect_Chain(instantiationNode))
        portAssociations = GetPortMapAspect(nodes.Get_Port_Map_Aspect_Chain(instantiationNode))

        return cls(instantiationNode, label, componentSymbol, genericAssociations, portAssociations)


@export
class EntityInstantiation(VHDLModel_EntityInstantiation, DOMMixin):
    def __init__(
        self,
        instantiationNode: Iir,
        label: str,
        entitySymbol: EntityInstantiationSymbol,
        architectureSymbol: ArchitectureSymbol = None,  # TODO: merge both symbols ?
        genericAssociations: Iterable[AssociationItem] = None,
        portAssociations: Iterable[AssociationItem] = None,
    ) -> None:
        super().__init__(label, entitySymbol, architectureSymbol, genericAssociations, portAssociations)
        DOMMixin.__init__(self, instantiationNode)

    @classmethod
    def parse(cls, instantiationNode: Iir, instantiatedUnit: Iir, label: str) -> "EntityInstantiation":
        from pyGHDL.dom._Translate import GetName, GetGenericMapAspect, GetPortMapAspect

        entityName = nodes.Get_Entity_Name(instantiatedUnit)
        entitySymbol = EntityInstantiationSymbol(entityName, GetName(entityName))

        architectureSymbol = None
        architectureId = nodes.Get_Architecture(instantiatedUnit)
        if architectureId != nodes.Null_Iir:
            architectureSymbol = ArchitectureSymbol(GetName(architectureId), entitySymbol)

        genericAssociations = GetGenericMapAspect(nodes.Get_Generic_Map_Aspect_Chain(instantiationNode))
        portAssociations = GetPortMapAspect(nodes.Get_Port_Map_Aspect_Chain(instantiationNode))

        return cls(instantiationNode, label, entitySymbol, architectureSymbol, genericAssociations, portAssociations)


@export
class ConfigurationInstantiation(VHDLModel_ConfigurationInstantiation, DOMMixin):
    def __init__(
        self,
        instantiationNode: Iir,
        label: str,
        configurationSymbol: ConfigurationInstantiationSymbol,
        genericAssociations: Iterable[AssociationItem] = None,
        portAssociations: Iterable[AssociationItem] = None,
    ) -> None:
        super().__init__(label, configurationSymbol, genericAssociations, portAssociations)
        DOMMixin.__init__(self, instantiationNode)

    @classmethod
    def parse(cls, instantiationNode: Iir, instantiatedUnit: Iir, label: str) -> "ConfigurationInstantiation":
        from pyGHDL.dom._Translate import GetName, GetGenericMapAspect, GetPortMapAspect

        configurationName = nodes.Get_Configuration_Name(instantiatedUnit)
        configurationSymbol = ConfigurationInstantiationSymbol(configurationName, GetName(configurationName))

        genericAssociations = GetGenericMapAspect(nodes.Get_Generic_Map_Aspect_Chain(instantiationNode))
        portAssociations = GetPortMapAspect(nodes.Get_Port_Map_Aspect_Chain(instantiationNode))

        return cls(instantiationNode, label, configurationSymbol, genericAssociations, portAssociations)


@export
class ConcurrentBlockStatement(VHDLModel_ConcurrentBlockStatement, DOMMixin):
    def __init__(
        self,
        blockNode: Iir,
        label: str,
        declaredItems: Iterable = None,
        statements: Iterable["ConcurrentStatement"] = None,
    ) -> None:
        super().__init__(label, None, declaredItems, statements)
        DOMMixin.__init__(self, blockNode)

    @classmethod
    def parse(cls, blockNode: Iir, label: str) -> "ConcurrentBlockStatement":
        from pyGHDL.dom._Translate import GetDeclaredItemsFromChainedNodes, GetConcurrentStatementsFromChainedNodes

        #        genericAssociations = GetGenericMapAspect(nodes.Get_Generic_Map_Aspect_Chain(instantiationNode))
        #        portAssociations = GetPortMapAspect(nodes.Get_Port_Map_Aspect_Chain(instantiationNode))

        declaredItems = GetDeclaredItemsFromChainedNodes(nodes.Get_Declaration_Chain(blockNode), "block", label)
        statements = GetConcurrentStatementsFromChainedNodes(
            nodes.Get_Concurrent_Statement_Chain(blockNode), "block", label
        )

        return cls(blockNode, label, declaredItems, statements)


@export
class ProcessStatement(VHDLModel_ProcessStatement, DOMMixin):
    def __init__(
        self,
        processNode: Iir,
        label: str = None,
        declaredItems: Iterable = None,
        statements: Iterable[SequentialStatement] = None,
        sensitivityList: Iterable[Symbol] = None,
    ) -> None:
        super().__init__(label, declaredItems, statements, sensitivityList)
        DOMMixin.__init__(self, processNode)

    @classmethod
    def parse(cls, processNode: Iir, label: str, hasSensitivityList: bool) -> "ProcessStatement":
        from pyGHDL.dom._Translate import (
            GetName,
            GetDeclaredItemsFromChainedNodes,
            GetSequentialStatementsFromChainedNodes,
        )

        sensitivityList = None
        if hasSensitivityList:
            sensitivityList = []
            for item in utils.list_iter(nodes.Get_Sensitivity_List(processNode)):
                sensitivityList.append(GetName(item))

        declaredItems = GetDeclaredItemsFromChainedNodes(nodes.Get_Declaration_Chain(processNode), "process", label)
        statements = GetSequentialStatementsFromChainedNodes(
            nodes.Get_Sequential_Statement_Chain(processNode), "process", label
        )

        return cls(processNode, label, declaredItems, statements, sensitivityList)


@export
class IfGenerateBranch(VHDLModel_IfGenerateBranch, DOMMixin):
    def __init__(
        self,
        branchNode: Iir,
        condition: ExpressionUnion,
        declaredItems: Iterable = None,
        statements: Iterable[ConcurrentStatement] = None,
        alternativeLabel: str = None,
    ) -> None:
        super().__init__(condition, declaredItems, statements, alternativeLabel)
        DOMMixin.__init__(self, branchNode)

    @classmethod
    def parse(cls, generateNode: Iir) -> "IfGenerateBranch":
        from pyGHDL.dom._Translate import (
            GetDeclaredItemsFromChainedNodes,
            GetConcurrentStatementsFromChainedNodes,
            GetExpressionFromNode,
        )

        condition = GetExpressionFromNode(nodes.Get_Condition(generateNode))
        body = nodes.Get_Generate_Statement_Body(generateNode)

        # TODO: alternative label
        # alternativeLabelId = nodes.Get_Alternative_Label(body)
        alternativeLabel = ""

        declarationChain = nodes.Get_Declaration_Chain(body)
        declaredItems = GetDeclaredItemsFromChainedNodes(declarationChain, "if-generate branch", alternativeLabel)

        statementChain = nodes.Get_Concurrent_Statement_Chain(body)
        statements = GetConcurrentStatementsFromChainedNodes(statementChain, "if-generate branch", alternativeLabel)

        return cls(generateNode, condition, declaredItems, statements, alternativeLabel)


@export
class ElsifGenerateBranch(VHDLModel_ElsifGenerateBranch, DOMMixin):
    def __init__(
        self,
        branchNode: Iir,
        condition: ExpressionUnion,
        declaredItems: Iterable = None,
        statements: Iterable[ConcurrentStatement] = None,
        alternativeLabel: str = None,
    ) -> None:
        super().__init__(condition, declaredItems, statements, alternativeLabel)
        DOMMixin.__init__(self, branchNode)

    @classmethod
    def parse(cls, generateNode: Iir, condition: Iir) -> "ElsifGenerateBranch":
        from pyGHDL.dom._Translate import (
            GetDeclaredItemsFromChainedNodes,
            GetConcurrentStatementsFromChainedNodes,
            GetExpressionFromNode,
        )

        condition = GetExpressionFromNode(condition)
        body = nodes.Get_Generate_Statement_Body(generateNode)

        # TODO: alternative label
        # alternativeLabelId = nodes.Get_Alternative_Label(body)
        alternativeLabel = ""

        declarationChain = nodes.Get_Declaration_Chain(body)
        declaredItems = GetDeclaredItemsFromChainedNodes(declarationChain, "elsif-generate branch", alternativeLabel)

        statementChain = nodes.Get_Concurrent_Statement_Chain(body)
        statements = GetConcurrentStatementsFromChainedNodes(statementChain, "elsif-generate branch", alternativeLabel)

        return cls(generateNode, condition, declaredItems, statements, alternativeLabel)


@export
class ElseGenerateBranch(VHDLModel_ElseGenerateBranch, DOMMixin):
    def __init__(
        self,
        branchNode: Iir,
        declaredItems: Iterable = None,
        statements: Iterable[ConcurrentStatement] = None,
        alternativeLabel: str = None,
    ) -> None:
        super().__init__(declaredItems, statements, alternativeLabel)
        DOMMixin.__init__(self, branchNode)

    @classmethod
    def parse(cls, generateNode: Iir) -> "ElseGenerateBranch":
        from pyGHDL.dom._Translate import (
            GetDeclaredItemsFromChainedNodes,
            GetConcurrentStatementsFromChainedNodes,
        )

        body = nodes.Get_Generate_Statement_Body(generateNode)

        # TODO: alternative label
        # alternativeLabelId = nodes.Get_Alternative_Label(body)
        alternativeLabel = ""

        declarationChain = nodes.Get_Declaration_Chain(body)
        declaredItems = GetDeclaredItemsFromChainedNodes(declarationChain, "else-generate branch", alternativeLabel)

        statementChain = nodes.Get_Concurrent_Statement_Chain(body)
        statements = GetConcurrentStatementsFromChainedNodes(statementChain, "else-generate branch", alternativeLabel)

        return cls(generateNode, declaredItems, statements, alternativeLabel)


@export
class IfGenerateStatement(VHDLModel_IfGenerateStatement, DOMMixin):
    def __init__(
        self,
        generateNode: Iir,
        label: str,
        ifBranch: IfGenerateBranch,
        elsifBranches: Iterable[ElsifGenerateBranch] = None,
        elseBranch: ElseGenerateBranch = None,
    ) -> None:
        super().__init__(label, ifBranch, elsifBranches, elseBranch)
        DOMMixin.__init__(self, generateNode)

    @classmethod
    def parse(cls, generateNode: Iir, label: str) -> "IfGenerateStatement":
        ifBranch = IfGenerateBranch.parse(generateNode)
        elsifBranches = []
        elseBranch = None
        # WORKAROUND: Python 3.8 syntax
        # elseClause = generateNode
        # while (elseClause := nodes.Get_Generate_Else_Clause(elseClause)) != nodes.Null_Iir:
        elseClause = nodes.Get_Generate_Else_Clause(generateNode)
        while elseClause != nodes.Null_Iir:
            condition = nodes.Get_Condition(elseClause)
            if condition != nodes.Null_Iir:
                elsifBranches.append(ElsifGenerateBranch.parse(elseClause, condition))
            else:
                elseBranch = ElseGenerateBranch.parse(elseClause)
                break

            elseClause = nodes.Get_Generate_Else_Clause(elseClause)

        return cls(generateNode, label, ifBranch, elsifBranches, elseBranch)


@export
class IndexedGenerateChoice(VHDLModel_IndexedGenerateChoice, DOMMixin):
    def __init__(self, node: Iir, expression: ExpressionUnion) -> None:
        super().__init__(expression)
        DOMMixin.__init__(self, node)


@export
class RangedGenerateChoice(VHDLModel_RangedGenerateChoice, DOMMixin):
    def __init__(self, node: Iir, rng: Range) -> None:
        super().__init__(rng)
        DOMMixin.__init__(self, node)


@export
class GenerateCase(VHDLModel_GenerateCase, DOMMixin):
    def __init__(
        self,
        node: Iir,
        choices: Iterable[ConcurrentChoice],
        declaredItems: Iterable = None,
        statements: Iterable[ConcurrentStatement] = None,
        alternativeLabel: str = None,
    ) -> None:
        super().__init__(choices, declaredItems, statements, alternativeLabel)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, caseNode: Iir, choices: Iterable[ConcurrentChoice]) -> "GenerateCase":
        from pyGHDL.dom._Translate import (
            GetDeclaredItemsFromChainedNodes,
            GetConcurrentStatementsFromChainedNodes,
        )

        body = nodes.Get_Associated_Block(caseNode)

        # TODO: alternative label
        # alternativeLabelId = nodes.Get_Alternative_Label(body)
        alternativeLabel = ""

        declarationChain = nodes.Get_Declaration_Chain(body)
        declaredItems = GetDeclaredItemsFromChainedNodes(declarationChain, "generate case", alternativeLabel)

        statementChain = nodes.Get_Concurrent_Statement_Chain(body)
        statements = GetConcurrentStatementsFromChainedNodes(statementChain, "generate case", alternativeLabel)

        return cls(caseNode, choices, declaredItems, statements, alternativeLabel)


@export
class OthersGenerateCase(VHDLModel_OthersGenerateCase, DOMMixin):
    def __init__(
        self,
        caseNode: Iir,
        declaredItems: Iterable = None,
        statements: Iterable[ConcurrentStatement] = None,
        alternativeLabel: str = None,
    ) -> None:
        super().__init__(declaredItems, statements, alternativeLabel)
        DOMMixin.__init__(self, caseNode)

    @classmethod
    def parse(cls, caseNode: Iir) -> "OthersGenerateCase":
        from pyGHDL.dom._Translate import (
            GetDeclaredItemsFromChainedNodes,
            GetConcurrentStatementsFromChainedNodes,
        )

        body = nodes.Get_Associated_Block(caseNode)

        # TODO: alternative label
        # alternativeLabelId = nodes.Get_Alternative_Label(body)
        alternativeLabel = ""

        declarationChain = nodes.Get_Declaration_Chain(body)
        declaredItems = GetDeclaredItemsFromChainedNodes(declarationChain, "case-generate others", alternativeLabel)

        statementChain = nodes.Get_Concurrent_Statement_Chain(body)
        statements = GetConcurrentStatementsFromChainedNodes(statementChain, "case-generate others", alternativeLabel)

        return cls(caseNode, declaredItems, statements, alternativeLabel)


@export
class CaseGenerateStatement(VHDLModel_CaseGenerateStatement, DOMMixin):
    def __init__(
        self,
        generateNode: Iir,
        label: str,
        expression: ExpressionUnion,
        cases: Iterable[ConcurrentCase],
    ) -> None:
        super().__init__(label, expression, cases)
        DOMMixin.__init__(self, generateNode)

    @classmethod
    def parse(cls, generateNode: Iir, label: str) -> "CaseGenerateStatement":
        from pyGHDL.dom._Utils import GetIirKindOfNode
        from pyGHDL.dom._Translate import (
            GetExpressionFromNode,
            GetRangeFromNode,
            GetName,
        )

        expression = GetExpressionFromNode(nodes.Get_Expression(generateNode))

        cases = []
        choices = None
        alternative = nodes.Get_Case_Statement_Alternative_Chain(generateNode)
        caseNode = alternative

        while alternative != nodes.Null_Iir:
            choiceKind = GetIirKindOfNode(alternative)
            sameAlternative = nodes.Get_Same_Alternative_Flag(alternative)

            if choiceKind in (
                nodes.Iir_Kind.Choice_By_Name,
                nodes.Iir_Kind.Choice_By_Expression,
            ):
                choiceExpression = GetExpressionFromNode(nodes.Get_Choice_Expression(alternative))

                choice = IndexedGenerateChoice(alternative, choiceExpression)
                if sameAlternative:
                    choices.append(choice)
                    alternative = nodes.Get_Chain(alternative)
                    continue
            elif choiceKind is nodes.Iir_Kind.Choice_By_Range:
                choiceRange = nodes.Get_Choice_Range(alternative)
                choiceRangeKind = GetIirKindOfNode(choiceRange)
                if choiceRangeKind == nodes.Iir_Kind.Range_Expression:
                    rng = GetRangeFromNode(choiceRange)
                elif choiceRangeKind in (
                    nodes.Iir_Kind.Attribute_Name,
                    nodes.Iir_Kind.Parenthesis_Name,
                ):
                    rng = GetName(choiceRange)
                else:
                    pos = Position.parse(alternative)
                    raise DOMException(
                        f"Unknown choice range kind '{choiceRangeKind.name}' in case...generate statement at line {pos.Line}."
                    )

                choice = RangedGenerateChoice(alternative, rng)
                if sameAlternative:
                    choices.append(choice)
                    alternative = nodes.Get_Chain(alternative)
                    continue
            elif choiceKind is nodes.Iir_Kind.Choice_By_Others:
                if choices is not None:
                    cases.append(GenerateCase.parse(alternative, choices))
                    choices = None
                cases.append(OthersGenerateCase.parse(alternative))
                alternative = nodes.Get_Chain(alternative)
                caseNode = alternative
                continue
            else:
                pos = Position.parse(alternative)
                raise DOMException(
                    f"Unknown choice kind '{choiceKind.name}' in case...generate statement at line {pos.Line}."
                )

            if choices is not None:
                cases.append(GenerateCase.parse(caseNode, choices))

            caseNode = alternative
            choices = [choice]

            alternative = nodes.Get_Chain(alternative)

        if choices is not None:
            cases.append(GenerateCase.parse(caseNode, choices))

        return cls(generateNode, label, expression, cases)


@export
class ForGenerateStatement(VHDLModel_ForGenerateStatement, DOMMixin):
    def __init__(
        self,
        generateNode: Iir,
        label: str,
        loopIndex: str,
        rng: Range,
        declaredItems: Iterable = None,
        statements: Iterable[ConcurrentStatement] = None,
    ) -> None:
        super().__init__(label, loopIndex, rng, declaredItems, statements)
        DOMMixin.__init__(self, generateNode)

    @classmethod
    def parse(cls, generateNode: Iir, label: str) -> "ForGenerateStatement":
        from pyGHDL.dom._Utils import GetIirKindOfNode, GetNameOfNode
        from pyGHDL.dom._Translate import (
            GetDeclaredItemsFromChainedNodes,
            GetConcurrentStatementsFromChainedNodes,
            GetRangeFromNode,
            GetName,
        )

        spec = nodes.Get_Parameter_Specification(generateNode)
        loopIndex = GetNameOfNode(spec)

        discreteRange = nodes.Get_Discrete_Range(spec)
        rangeKind = GetIirKindOfNode(discreteRange)
        if rangeKind == nodes.Iir_Kind.Range_Expression:
            rng = GetRangeFromNode(discreteRange)
        elif rangeKind in (
            nodes.Iir_Kind.Attribute_Name,
            nodes.Iir_Kind.Parenthesis_Name,
        ):
            rng = GetName(discreteRange)
        else:
            pos = Position.parse(generateNode)
            raise DOMException(
                f"Unknown discrete range kind '{rangeKind.name}' in for...generate statement at line {pos.Line}."
            )

        body = nodes.Get_Generate_Statement_Body(generateNode)
        declarationChain = nodes.Get_Declaration_Chain(body)
        declaredItems = GetDeclaredItemsFromChainedNodes(declarationChain, "for-generate", label)

        statementChain = nodes.Get_Concurrent_Statement_Chain(body)
        statements = GetConcurrentStatementsFromChainedNodes(statementChain, "for-generate", label)

        return cls(generateNode, label, loopIndex, rng, declaredItems, statements)


@export
class WaveformElement(VHDLModel_WaveformElement, DOMMixin):
    def __init__(self, waveNode: Iir, expression: ExpressionUnion, after: ExpressionUnion) -> None:
        super().__init__(expression, after)
        DOMMixin.__init__(self, waveNode)

    @classmethod
    def parse(cls, waveNode: Iir):
        from pyGHDL.dom._Translate import GetExpressionFromNode

        value = GetExpressionFromNode(nodes.Get_We_Value(waveNode))

        timeNode = nodes.Get_Time(waveNode)
        if timeNode is nodes.Null_Iir:
            time = None
        else:
            time = GetExpressionFromNode(timeNode)

        return cls(waveNode, value, time)


@export
class ConcurrentSimpleSignalAssignment(VHDLModel_ConcurrentSimpleSignalAssignment, DOMMixin):
    def __init__(
        self,
        assignmentNode: Iir,
        label: str,
        target: Symbol,
        waveform: Iterable[WaveformElement],
    ) -> None:
        super().__init__(label, target, waveform)
        DOMMixin.__init__(self, assignmentNode)

    @classmethod
    def parse(cls, assignmentNode: Iir, label: str) -> "ConcurrentSimpleSignalAssignment":
        from pyGHDL.dom._Translate import GetName

        target = nodes.Get_Target(assignmentNode)
        targetName = GetName(target)

        waveform = []
        for wave in utils.chain_iter(nodes.Get_Waveform_Chain(assignmentNode)):
            waveform.append(WaveformElement.parse(wave))

        return cls(assignmentNode, label, targetName, waveform)


@export
class ConcurrentProcedureCall(VHDLModel_ConcurrentProcedureCall, DOMMixin):
    def __init__(
        self,
        callNode: Iir,
        label: str,
        procedureName: Symbol,
        parameterMappings: Iterable,
    ) -> None:
        super().__init__(label, procedureName, parameterMappings)
        DOMMixin.__init__(self, callNode)

    @classmethod
    def parse(cls, concurrentCallNode: Iir, label: str) -> "ConcurrentProcedureCall":
        from pyGHDL.dom._Translate import GetName, GetParameterMapAspect

        callNode = nodes.Get_Procedure_Call(concurrentCallNode)

        prefix = nodes.Get_Prefix(callNode)
        procedureName = GetName(prefix)
        parameterAssociations = GetParameterMapAspect(nodes.Get_Parameter_Association_Chain(callNode))

        return cls(concurrentCallNode, label, procedureName, parameterAssociations)


@export
class ConcurrentAssertStatement(VHDLModel_ConcurrentAssertStatement, DOMMixin):
    def __init__(
        self,
        assertNode: Iir,
        condition: ExpressionUnion,
        message: ExpressionUnion = None,
        severity: ExpressionUnion = None,
        label: str = None,
    ) -> None:
        super().__init__(condition, message, severity, label)
        DOMMixin.__init__(self, assertNode)

    @classmethod
    def parse(cls, assertNode: Iir, label: str) -> "ConcurrentAssertStatement":
        from pyGHDL.dom._Translate import GetExpressionFromNode

        # FIXME: how to get the condition?
        # assertNode is a Psl_Assert_Directive
        condition = None  # GetExpressionFromNode(nodes.Get_Assertion_Condition(assertNode))
        messageNode = nodes.Get_Report_Expression(assertNode)
        message = None if messageNode is nodes.Null_Iir else GetExpressionFromNode(messageNode)
        severityNode = nodes.Get_Severity_Expression(assertNode)
        severity = None if severityNode is nodes.Null_Iir else GetExpressionFromNode(severityNode)

        return cls(assertNode, condition, message, severity, label)



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Package module:   DOM: Interface items (e.g. generic or port)
#
# License:
# ============================================================================
#  Copyright (C) 2019-2022 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================
"""
This module implements derived type classes from :mod:`pyVHDLModel.Type`.
"""
from typing import List, Union, Iterator, Tuple, Iterable

from pyGHDL.dom.Name import SimpleName
from pyTooling.Decorators import export

from pyVHDLModel.Name import Name
from pyVHDLModel.Symbol import Symbol
from pyVHDLModel.Type import Subtype as VHDLModel_Subtype
from pyVHDLModel.Type import AnonymousType as VHDLModel_AnonymousType
from pyVHDLModel.Type import EnumeratedType as VHDLModel_EnumeratedType
from pyVHDLModel.Type import IntegerType as VHDLModel_IntegerType
from pyVHDLModel.Type import PhysicalType as VHDLModel_PhysicalType
from pyVHDLModel.Type import ArrayType as VHDLModel_ArrayType
from pyVHDLModel.Type import RecordTypeElement as VHDLModel_RecordTypeElement
from pyVHDLModel.Type import RecordType as VHDLModel_RecordType
from pyVHDLModel.Type import ProtectedType as VHDLModel_ProtectedType
from pyVHDLModel.Type import ProtectedTypeBody as VHDLModel_ProtectedTypeBody
from pyVHDLModel.Type import AccessType as VHDLModel_AccessType
from pyVHDLModel.Type import FileType as VHDLModel_FileType

from pyGHDL.libghdl import utils
from pyGHDL.libghdl._types import Iir
from pyGHDL.libghdl.vhdl import nodes, flists
from pyGHDL.dom import DOMMixin, DOMException, Position
from pyGHDL.dom.Symbol import SimpleSubtypeSymbol
from pyGHDL.dom.Literal import EnumerationLiteral, PhysicalIntegerLiteral
from pyGHDL.dom.Range import Range
from pyGHDL.dom.Subprogram import Function, Procedure


@export
class IncompleteType(VHDLModel_AnonymousType, DOMMixin):
    def __init__(self, node: Iir, identifier: str) -> None:
        super().__init__(identifier, parent=None)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, node: Iir) -> "IncompleteType":
        from pyGHDL.dom._Utils import GetNameOfNode

        name = GetNameOfNode(node)

        return cls(node, name)


@export
class EnumeratedType(VHDLModel_EnumeratedType, DOMMixin):
    """
    Represents an *enumerated type*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Type.EnumeratedType`.

    .. admonition:: Example

       .. code-block:: VHDL

          type integer is (lit_1, lit2, ...);
    """

    def __init__(self, node: Iir, identifier: str, literals: List[EnumerationLiteral]) -> None:
        super().__init__(identifier, literals, None)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, typeName: str, typeDefinitionNode: Iir) -> "EnumeratedType":
        """
        Parses an *enumerated type* IIR and returns an :class:`~pyVHDLModel.Type.EnumeratedType` instance.

        :param typeName:           The identifier of the type.
        :param typeDefinitionNode: The IIR node to parse.
        :return:                   The enumerated type instance.
        """
        literals = []
        enumerationLiterals = nodes.Get_Enumeration_Literal_List(typeDefinitionNode)
        for enumerationLiteral in utils.flist_iter(enumerationLiterals):
            literal = EnumerationLiteral.parse(enumerationLiteral)
            literals.append(literal)

        return cls(typeDefinitionNode, typeName, literals)


@export
class IntegerType(VHDLModel_IntegerType, DOMMixin):
    """
    Represents an *integer type*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Type.IntegerType`.

    .. admonition:: Example

       .. code-block:: VHDL

          type integer is range -2147483648 to 2147483647;
    """

    def __init__(self, node: Iir, typeName: str, rng: Union[Range, "Name"]) -> None:
        super().__init__(typeName, rng, None)
        DOMMixin.__init__(self, node)


@export
class PhysicalType(VHDLModel_PhysicalType, DOMMixin):
    """
    Represents a *physical type*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Type.PhysicalType`.

    .. admonition:: Example

       .. code-block:: VHDL

          type time is range integer'low to integer'high units
            fs;
            ps = 1000 fs;
            -- ...
          end units;
    """

    def __init__(
        self,
        node: Iir,
        typeName: str,
        rng: Union[Range, Name],
        primaryUnit: str,
        units: List[Tuple[str, PhysicalIntegerLiteral]],
    ) -> None:
        super().__init__(typeName, rng, primaryUnit, units, None)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, typeName: str, typeDefinitionNode: Iir) -> "PhysicalType":
        """
        Parses an *physical type* IIR and returns an :class:`~pyVHDLModel.Type.PhysicalType` instance.

        :param typeName:           The identifier of the type.
        :param typeDefinitionNode: The IIR node to parse.
        :return:                   The physical type instance.
        """
        from pyGHDL.dom._Utils import GetIirKindOfNode, GetNameOfNode
        from pyGHDL.dom._Translate import GetRangeFromNode, GetName

        rangeConstraint = nodes.Get_Range_Constraint(typeDefinitionNode)
        rangeKind = GetIirKindOfNode(rangeConstraint)
        if rangeKind == nodes.Iir_Kind.Range_Expression:
            rng = GetRangeFromNode(rangeConstraint)
        elif rangeKind in (
            nodes.Iir_Kind.Attribute_Name,
            nodes.Iir_Kind.Parenthesis_Name,
        ):
            rng = GetName(rangeConstraint)
        else:
            pos = Position.parse(typeDefinitionNode)
            raise DOMException(f"Unknown range kind '{rangeKind.name}' in physical type definition at line {pos.Line}.")

        primaryUnit = nodes.Get_Primary_Unit(typeDefinitionNode)
        primaryUnitName = GetNameOfNode(primaryUnit)

        units = []
        for secondaryUnit in utils.chain_iter(nodes.Get_Unit_Chain(typeDefinitionNode)):
            secondaryUnitName = GetNameOfNode(secondaryUnit)
            if secondaryUnit == primaryUnit:
                continue

            physicalLiteral = PhysicalIntegerLiteral.parse(nodes.Get_Physical_Literal(secondaryUnit))

            units.append((secondaryUnitName, physicalLiteral))

        return cls(typeDefinitionNode, typeName, rng, primaryUnitName, units)


@export
class ArrayType(VHDLModel_ArrayType, DOMMixin):
    """
    Represents an *array type*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Type.ArrayType`.

    .. admonition:: Example

       .. code-block:: VHDL

          type bit_vector is array(natural range <>) of bit;
    """

    def __init__(self, node: Iir, identifier: str, indices: List, elementSubtype: Symbol) -> None:
        super().__init__(identifier, indices, elementSubtype, None)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, typeName: str, typeDefinitionNode: Iir) -> "ArrayType":
        """
        Parses an *array type* IIR and returns an :class:`~pyVHDLModel.Type.ArrayType` instance.

        :param typeName:           The identifier of the type.
        :param typeDefinitionNode: The IIR node to parse.
        :return:                   The array type instance.
        """
        from pyGHDL.dom._Utils import GetIirKindOfNode
        from pyGHDL.dom._Translate import (
            GetSimpleTypeFromNode,
            GetSubtypeIndicationFromIndicationNode,
        )

        indices = []
        indexDefinitions = nodes.Get_Index_Subtype_Definition_List(typeDefinitionNode)
        for index in utils.flist_iter(indexDefinitions):
            indexKind = GetIirKindOfNode(index)
            if indexKind == nodes.Iir_Kind.Simple_Name:
                indexSubtype = GetSimpleTypeFromNode(index)
                indices.append(indexSubtype)
            else:
                raise DOMException(
                    f"Unknown kind '{indexKind.name}' for an index in the array definition of `{typeName}`."
                )

        elementSubtypeIndication = nodes.Get_Element_Subtype_Indication(typeDefinitionNode)
        elementSubtype = GetSubtypeIndicationFromIndicationNode(elementSubtypeIndication, "array declaration", typeName)

        return cls(typeDefinitionNode, typeName, indices, elementSubtype)


@export
class RecordTypeElement(VHDLModel_RecordTypeElement, DOMMixin):
    """
    Represents a *record element*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Type.RecordTypeElement`.

    .. admonition:: Example

       .. code-block:: VHDL

          -- type pt is record
            element : std_logic;
            -- ...
          -- end record;
    """

    def __init__(self, node: Iir, identifiers: List[str], subtype: Symbol) -> None:
        super().__init__(identifiers, subtype, None)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, elementDeclarationNode: Iir, furtherIdentifiers: Iterable[str] = None) -> "RecordTypeElement":
        """
        Parses a *record element* IIR and returns an :class:`~pyVHDLModel.Type.RecordTypeElement` instance.

        :param elementDeclarationNode: The IIR node to parse.
        :param furtherIdentifiers:     The list of record element identifiers.
        :return:                       The record element instance.
        """
        from pyGHDL.dom._Utils import GetNameOfNode
        from pyGHDL.dom._Translate import GetSubtypeIndicationFromNode

        elementName = GetNameOfNode(elementDeclarationNode)
        elementType = GetSubtypeIndicationFromNode(elementDeclarationNode, "record element", elementName)

        identifiers = [elementName]
        if furtherIdentifiers is not None:
            identifiers.extend(furtherIdentifiers)

        return cls(elementDeclarationNode, identifiers, elementType)


@export
class RecordType(VHDLModel_RecordType, DOMMixin):
    """
    Represents a *record type*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Type.RecordType`.

    .. admonition:: Example

       .. code-block:: VHDL

          type pt is record
            -- elements
          end record;
    """

    def __init__(self, node: Iir, identifier: str, elements: List[RecordTypeElement] = None) -> None:
        super().__init__(identifier, elements, None)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, typeName: str, typeDefinitionNode: Iir) -> "RecordType":
        """
        Parses a *record type* IIR and returns an :class:`~pyVHDLModel.Type.RecordType` instance.

        :param typeName:           The identifier of the type.
        :param typeDefinitionNode: The IIR node to parse.
        :return:                   The record type instance.
        """
        from pyGHDL.dom._Utils import GetNameOfNode

        elements = []
        elementDeclarations = nodes.Get_Elements_Declaration_List(typeDefinitionNode)

        furtherIdentifiers = []
        elementCount = flists.Flast(elementDeclarations) + 1
        index = 0
        while index < elementCount:
            elementDeclaration = flists.Get_Nth_Element(elementDeclarations, index)

            # Lookahead for elements with multiple identifiers at once
            if nodes.Get_Has_Identifier_List(elementDeclaration):
                index += 1
                while index < elementCount:
                    nextNode: Iir = flists.Get_Nth_Element(elementDeclarations, index)
                    # Consecutive identifiers are found, if the subtype indication is Null
                    if nodes.Get_Subtype_Indication(nextNode) == nodes.Null_Iir:
                        furtherIdentifiers.append(GetNameOfNode(nextNode))
                    else:
                        break
                    index += 1

                    # The last consecutive identifiers has no Identifier_List flag
                    if not nodes.Get_Has_Identifier_List(nextNode):
                        break
            else:
                index += 1

            element = RecordTypeElement.parse(elementDeclaration, furtherIdentifiers)
            elements.append(element)
            furtherIdentifiers.clear()

        return cls(typeDefinitionNode, typeName, elements)


@export
class ProtectedType(VHDLModel_ProtectedType, DOMMixin):
    """
    Represents a *protected type*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Type.ProtectedType`.

    .. admonition:: Example

       .. code-block:: VHDL

          type pt is protected
            -- public interface
          end protected;
    """

    def __init__(self, node: Iir, identifier: str, methods: Union[List, Iterator] = None) -> None:
        super().__init__(identifier, methods, None)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, typeName: str, typeDefinitionNode: Iir) -> "ProtectedType":
        """
        Parses a *protected type* IIR and returns an :class:`~pyVHDLModel.Type.ProtectedType` instance.

        :param typeName:           The identifier of the type.
        :param typeDefinitionNode: The IIR node to parse.
        :return:                   The protected type instance.
        """
        from pyGHDL.dom._Utils import GetIirKindOfNode

        # FIXME: change this to a generator
        methods = []
        for item in utils.chain_iter(nodes.Get_Declaration_Chain(typeDefinitionNode)):
            kind = GetIirKindOfNode(item)
            if kind == nodes.Iir_Kind.Function_Declaration:
                methods.append(Function.parse(item))
            elif kind == nodes.Iir_Kind.Procedure_Declaration:
                methods.append(Procedure.parse(item))

        return cls(typeDefinitionNode, typeName, methods)


@export
class ProtectedTypeBody(VHDLModel_ProtectedTypeBody, DOMMixin):
    """
    Represents a *protected type body*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Type.ProtectedTypeBody`.

    .. admonition:: Example

       .. code-block:: VHDL

          type pt is protected body
            -- implementations
          end protected body;
    """

    def __init__(self, node: Iir, identifier: str, declaredItems: Union[List, Iterator] = None) -> None:
        super().__init__(identifier, declaredItems, None)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, protectedBodyNode: Iir) -> "ProtectedTypeBody":
        """
        Parses a *protected type body* IIR and returns an :class:`~pyVHDLModel.Type.ProtectedTypeBody` instance.

        :param protectedBodyNode: The IIR node to parse.
        :return:                  The protected type body instance.
        """
        from pyGHDL.dom._Utils import GetNameOfNode
        from pyGHDL.dom._Translate import GetDeclaredItemsFromChainedNodes

        typeName = GetNameOfNode(protectedBodyNode)
        declaredItems = GetDeclaredItemsFromChainedNodes(
            nodes.Get_Declaration_Chain(protectedBodyNode),
            "protected type body",
            typeName,
        )

        return cls(protectedBodyNode, typeName, declaredItems)


@export
class AccessType(VHDLModel_AccessType, DOMMixin):
    """
    Represents an *access type*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Type.AccessType`.

    .. admonition:: Example

       .. code-block:: VHDL

          type line is access string;
    """

    def __init__(self, node: Iir, identifier: str, designatedSubtype: Symbol) -> None:
        super().__init__(identifier, designatedSubtype, None)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, typeName: str, typeDefinitionNode: Iir) -> "AccessType":
        """
        Parses an *access type* IIR and returns an :class:`~pyVHDLModel.Type.AccessType` instance.

        :param typeName:           The identifier of the type.
        :param typeDefinitionNode: The IIR node to parse.
        :return:                   The access type instance.
        """
        from pyGHDL.dom._Translate import GetSubtypeIndicationFromIndicationNode

        designatedSubtypeIndication = nodes.Get_Designated_Subtype_Indication(typeDefinitionNode)
        designatedSubtype = GetSubtypeIndicationFromIndicationNode(designatedSubtypeIndication, "access type", typeName)

        return cls(typeDefinitionNode, typeName, designatedSubtype)


@export
class FileType(VHDLModel_FileType, DOMMixin):
    """
    Represents a *file type*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Type.FileType`.

    .. admonition:: Example

       .. code-block:: VHDL

          type text is file of string;
    """

    def __init__(self, node: Iir, identifier: str, designatedSubtype: Symbol) -> None:
        super().__init__(identifier, designatedSubtype, None)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, typeName: str, typeDefinitionNode: Iir) -> "FileType":
        """
        Parses a *file type* IIR and returns an :class:`~pyVHDLModel.Type.FileType` instance.

        :param typeName:           The identifier of the type.
        :param typeDefinitionNode: The IIR node to parse.
        :return:                   The file type instance.
        """
        from pyGHDL.dom._Utils import GetNameOfNode

        designatedSubtypeMark = nodes.Get_File_Type_Mark(typeDefinitionNode)
        designatedSubtypeName = GetNameOfNode(designatedSubtypeMark)
        designatedSubtype = SimpleSubtypeSymbol(
            typeDefinitionNode, SimpleName(designatedSubtypeMark, designatedSubtypeName)
        )

        return cls(typeDefinitionNode, typeName, designatedSubtype)


@export
class Subtype(VHDLModel_Subtype, DOMMixin):
    """
    Represents a *subtype*.

    This class implements a :mod:`pyGHDL.dom` object derived from :class:`pyVHDLModel.Type.Subtype`.
    """

    def __init__(self, node: Iir, subtypeName: str, symbol: Symbol) -> None:
        super().__init__(subtypeName, symbol, None)
        DOMMixin.__init__(self, node)



#!/usr/bin/env python3

"""Like pnodes but output for Rust."""

import re
import sys
from textwrap import dedent

try:
    import scripts.pnodes as pnodes
except:
    import pnodes


def print_enum(name, vals):
    if len(vals) < 256:
        repr = 8
    elif len(vals) < 65536:
        repr = 16
    else:
        repr = 32
    print(f"#[repr(u{repr})]")
    print(f"pub enum {name} {{")
    for k in vals:
        print(f"    {k},")
    print(f"}}");

    print(f"")
    print(f"impl {name} {{")
    print(f"    pub const VALUES: [Self; {len(vals)}] = [")
    for k in vals:
        print(f"        Self::{k},")
    print(f"    ];");
    print(f"")
    print(f"    pub const IMAGES: [&'static str; {len(vals)}] = [")
    for k in vals:
        print(f"        \"{k.lower()}\",")
    print(f"    ];");
    print(f"}}")

def print_file_header(includeIntEnumUnique=True, includeBindToLibGHDL=True):
    print(dedent("""\
            # Auto generated Python source file from Ada sources
            # Call 'make' in 'src/vhdl' to regenerate:
            #
        """) + "{sysImports}from pyTooling.Decorators import export\n{moduleImports}".format(
            sysImports = "from enum import IntEnum, unique\n" if includeIntEnumUnique else "",
            moduleImports = "\nfrom pyGHDL.libghdl._decorator import BindToLibGHDL\n" if includeBindToLibGHDL else "",
        )
    )


def do_class_kinds():
    typ = "Kind"
    print(f"#[derive(Copy, Clone, PartialEq, PartialOrd)]")
    print_enum(typ, pnodes.kinds)
    print(f"impl {typ} {{")
    for k, v in pnodes.kinds_ranges.items():
        print(f"    fn is_{k.lower()}(self: Self) -> bool {{")
        print(f"        self >= {typ}::{v[0]} && self <= {typ}::{v[-1]}")
        print(f"    }}")
        print()
    print(f"}}")

def common_subprg_import_header(pfx, nname):
    print(f'extern "C" {{')
    print(f'    #[link_name = "{pfx}__create_{nname}"]')
    print(f"    fn create(k: Kind) -> Node;")
    print()
    print(f'    #[link_name = "{pfx}__get_kind"]')
    print(f"    fn get_kind(n: Node) -> Kind;")
    print()
    print(f'    #[link_name = "{pfx}__get_location"]')
    print(f"    fn get_location(n: Node) -> Location;")
    print()
    print(f'    #[link_name = "{pfx}__set_location"]')
    print(f"    fn set_location(n: Node, loc: Location);")
    print()

def common_subprg_import(pnodes, pfx, nname, convert):
    common_subprg_import_header(pfx, nname)
    for k in pnodes.funcs:
        rtype = convert(k.rtype)
        name = k.name.lower()
        print(f'    #[link_name = "{pfx}__get_{name}"]')
        print(f"    fn get_{name}(n: Node) -> {rtype};")
        print()
        print(f'    #[link_name = "{pfx}__set_{name}"]')
        print(f"    fn set_{name}(n: Node, v: {rtype});")
        print()


def common_subprg_impl_header():
    print()
    print(f'impl Node {{')
    print(f'    pub const NULL: Self = Node(0);')
    print()
    print(f'    pub fn new(k: Kind) -> Self {{')
    print(f'        unsafe {{ create(k) }}')
    print(f'    }}')
    print()
    print(f'    pub fn kind(self: Self) -> Kind {{')
    print(f'        unsafe {{ get_kind(self) }}')
    print(f'    }}')
    print()
    print(f'    pub fn location(self: Self) -> Location {{')
    print(f'        unsafe {{ get_location(self) }}')
    print(f'    }}')
    print()
    print(f'    pub fn set_location(self: Self, loc: Location) {{')
    print(f'        unsafe {{ set_location(self, loc) }}')
    print(f'    }}')
    print()

def common_subprg_impl(pnodes, convert):
    namemap = {'type': 'typed'}
    common_subprg_impl_header()
    for k in pnodes.funcs:
        rtype = convert(k.rtype)

        name = k.name.lower()
        rname = namemap.get(name, name)
        print(f'    pub fn {rname}(self: Self) -> {rtype} {{')
        print(f'        unsafe {{ get_{name}(self) }}')
        print(f'    }}')
        print()
        print(f'    pub fn set_{rname}(self: Self, v : {rtype}) {{')
        print(f'        unsafe {{ set_{name}(self, v); }}')
        print(f'    }}')
        print()
    print(f"}}")

def convert_vhdl_type_name(rtype):
    typmap = {'TokenType': 'Tok',
              'Boolean' : 'bool',
              'Int32': 'i32',
              'Int64': 'i64',
              'Fp64': 'f64',
              'Iir': 'Node',
              }
    # Don't use the Iir_* subtypes (as they are not described).
    rtype = rtype.replace("Iir_", "")
    rtype = rtype.replace("_", "")
    rtype = rtype if not rtype.startswith("Kind_") else "Iir"
    # Exceptions...
    rtype = typmap.get(rtype, rtype)
    return rtype

def do_vhdl_subprg():
    pfx = "vhdl__nodes"
    print(f'#[repr(transparent)]')
    print(f'#[derive(Copy, Clone, PartialEq)]')
    print(f'pub struct Node(u32);')
    print()
    print(f'#[repr(transparent)]')
    print(f'#[derive(Copy, Clone, PartialEq)]')
    print(f'pub struct Flist(u32);')
#    print(f'type Iir = u32;')
#    print(f'type FileChecksumId = u32;')
#    print(f'type TimeStampId = u32;')
#    print(f'type SourceFileEntry = u32;')
#    print(f'type DateType = u32;')
#    print(f'type NameId = u32;')
#    print(f'type SourcePtr = u32;')
#    print(f'type String8Id = u32;')
    print(f'type PSLNode = u32;')
    print(f'type PSLNFA = u32;')
    print(f'type Tok = u8;') # FIXME
    print(f'type List = u32;')
    print(f'type Index32 = i32;')
    print()
    print(f'#[repr(u8)]')
    print(f'pub enum TriStateType {{')
    print(f'    Unknown,')
    print(f'    False,')
    print(f'    True,')
    print(f'}}')
    print()
    print(f'#[repr(u8)]')
    print(f'pub enum DirectionType {{')
    print(f'    To,')
    print(f'    Downto,')
    print(f'}}')
    print()

    common_subprg_import(pnodes, pfx, "iir", convert_vhdl_type_name)
    print('    #[link_name = "vhdl__flists__create_flist"]')
    print('    fn create_flist(len: u32) -> Flist;')
    print()
    print('    #[link_name = "vhdl__flists__set_nth_element"]')
    print('    fn set_nth_element(flist: Flist, idx: u32, el: Node);')
    print()
    print('    #[link_name = "vhdl__flists__get_nth_element"]')
    print('    fn get_nth_element(flist: Flist, idx: u32) -> Node;')
    print()
    print(f"}}")

    common_subprg_impl(pnodes, convert_vhdl_type_name)

    print('impl Flist {')
    print('    pub fn new(len: u32) -> Self {')
    print('        unsafe { create_flist(len) }')
    print('    }')
    print()
    print('    pub fn set(self: Self, idx: u32, el: Node) {')
    print('        unsafe { set_nth_element(self, idx, el); }')
    print('    }')
    print('}')


def do_class_types():
    print_enum("types", pnodes.get_types())


def do_has_subprg():
    print()
    for f in pnodes.funcs:
        print(dedent(f"""
            @export
            @BindToLibGHDL("vhdl__nodes_meta__has_{f.name.lower()}")
            def Has_{f.name}(kind: IirKind) -> bool:
                \"\"\"\"\"\"
            """)
        )


def do_class_field_attributes():
    print_enum("Attr", ["ANone" if a == "None" else a for a in pnodes.get_attributes()])


def do_class_fields():
    print_enum("fields", [f.name for f in pnodes.funcs])


def read_spec_enum(type_name, prefix, class_name):
    """Read an enumeration declaration from kind file."""
    toks = pnodes.read_enum(pnodes.kind_file, type_name, prefix)
    print_enum(class_name, toks)


def do_vhdl_nodes():
    """Convert vhdl-nodes.ads"""
    print("#![allow(non_camel_case_types, dead_code)]")
    print("use crate::types::*;")
    print("use crate::files_map::{Location, SourceFileEntry};")
    print("use crate::NameId;")
    print()
    do_class_kinds()
    read_spec_enum("Iir_Mode", "Iir_", "Mode")
    read_spec_enum("Scalar_Size", "", "ScalarSize")
    read_spec_enum("Iir_Staticness", "", "Staticness")
    read_spec_enum("Iir_Constraint", "", "Iir_Constraint")
    read_spec_enum("Iir_Delay_Mechanism", "Iir_", "DelayMechanism")
    read_spec_enum("Iir_Pure_State", "", "PureState")
    read_spec_enum("Iir_All_Sensitized", "", "AllSensitized")
    read_spec_enum("Iir_Signal_Kind", "Iir_", "SignalKind")
    read_spec_enum("Iir_Constraint", "", "Constraint")
    read_spec_enum("Iir_Force_Mode", "Iir_", "ForceMode")
    read_spec_enum("Date_State_Type", "Date_", "DateStateType")
    read_spec_enum("Number_Base_Type", "", "NumberBaseType")
    read_spec_enum("Iir_Predefined_Functions", "Iir_Predefined_", "PredefinedFunctions")
    do_vhdl_subprg()


def convert_verilog_type_name(rtype):
    typmap = {'Boolean' : 'bool',
              'Int32': 'i32',
              'Int64': 'i64',
              'Fp64': 'f64',
              'Uns32': 'u32',
              }
    # Rename '_' and 'Type'
    rtype = rtype.replace("_", "").replace("Type","")
    # Exceptions...
    rtype = typmap.get(rtype, rtype)
    return rtype


def do_verilog_subprg():
    pfx = "verilog__nodes"
    print()
    common_subprg_import(pnodes, pfx, "node", convert_verilog_type_name)
    print(f"}}")

    common_subprg_impl(pnodes, convert_verilog_type_name)


def do_verilog_nodes():
    """Convert verilog-nodes.ads"""
    print("#![allow(non_camel_case_types, dead_code)]")
    print()
    print("use crate::types::*;")
    print("use crate::verilog::types::*;")
    print("use crate::files_map::Location;")
    print("use crate::NameId;")
    print()
    print("#[repr(transparent)]")
    print("#[derive(Copy, Clone, PartialEq)]")
    print("pub struct Node(u32);")
    print()
    do_class_kinds()
    read_spec_enum("Violation_Type", "Violation_", "Violation")
    read_spec_enum("Visibility_Type", "Visibility_", "Visibility")
    read_spec_enum("DPI_Spec_Type", "DPI_", "DPISpec")
    read_spec_enum("Edge_Type", "Edge_", "Edge")
    read_spec_enum("Base_Type", "Base_", "Base")
    read_spec_enum("Lifetime_Type", "Life_", "Lifetime")
    read_spec_enum("Join_Type", "Join_", "Join")
    read_spec_enum("Udp_Symbol", "", "UdpSymbol")
    read_spec_enum("Polarity_Type", "Polarity_", "Polarity")
    read_spec_enum("Udp_Kind", "Udp_", "UdpKind")
    read_spec_enum("Conv_Ops", "Convop_", "ConvOps")
    read_spec_enum("Binary_Ops", "Binop_", "BinaryOps")
    read_spec_enum("Unary_Ops", "Unop_", "UnaryOps")
    print("pub type SysTfId = u32;")
    do_verilog_subprg()


def do_std_names():
    res = pnodes.read_std_names()
    print('#![allow(dead_code)]')
    print('use crate::NameId;')
    for n, v in res:
#        # Avoid clash with Python names
#        if n in ["False", "True", "None"]:
#            n = "N" + n
        print(f"pub const {n.upper()}: NameId = NameId({v});")


def do_errorout():
    print("#![allow(dead_code)]")
    toks = pnodes.read_enum(
        "../errorout.ads",
        "Msgid_Type",
        "(Msgid|Warnid)_",
        g=lambda m: m.group(1) + "_" + m.group(2),
    )
    first_warnid = None
    for v, t in enumerate(toks):
        if t.startswith("Msgid_"):
            print(f"pub const {t.upper()}: u8 = {v};")
        elif first_warnid is None:
            first_warnid = t
            print(f"pub const MSGID_FIRST_WARNID: u8 = {v};")

    print("")
    # Remove prefix, remove '_'
    pfx="Warnid_"
    warn_toks = [t[len(pfx):].replace('_', '')
                 for t in toks if t.startswith(pfx)]
    print_enum("Warnid", warn_toks)
    print(f"pub const WARNID_USIZE: usize = {len(warn_toks)};")


pnodes.actions.update(
    {
        "vhdl-nodes": do_vhdl_nodes,
        "errorout": do_errorout,
        "verilog-nodes": do_verilog_nodes,
        "std_names": do_std_names,
        "class-kinds": do_class_kinds,
    }
)


def _generateCLIParser():
    return pnodes._generateCLIParser()


if __name__ == "__main__":
    pnodes.main()



# =============================================================================
#               ____ _   _ ____  _       _ _ _           _         _ _
#  _ __  _   _ / ___| | | |  _ \| |     | (_) |__   __ _| |__   __| | |
# | '_ \| | | | |  _| |_| | | | | |     | | | '_ \ / _` | '_ \ / _` | |
# | |_) | |_| | |_| |  _  | |_| | |___ _| | | |_) | (_| | | | | (_| | |
# | .__/ \__, |\____|_| |_|____/|_____(_)_|_|_.__/ \__, |_| |_|\__,_|_|
# |_|    |___/                                     |___/
# =============================================================================
# Authors:
#   Tristan Gingold
#   Patrick Lehmann
#
# Package module:   Python binding and low-level API for shared library 'libghdl'.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================

from ctypes import c_int32, c_char_p, c_bool, c_uint32

from pyTooling.Decorators import export

from pyGHDL.libghdl import libghdl, ENCODING
from pyGHDL.libghdl._decorator import BindToLibGHDL
from pyGHDL.libghdl._types import SourceFileEntry


# @export
@BindToLibGHDL("files_map__editor__replace_text_ptr")
def _Replace_Text(
    File: SourceFileEntry,
    Start_Line: int,
    Start_Offset: int,
    End_Line: int,
    End_Offset: int,
    Text_Pointer: c_char_p,
    Text_Length: int,
) -> bool:
    """Replace [START; END) by TEXT.

    :param File:         File where to replace a text section.
    :param Start_Line:
    :param Start_Offset:
    :param End_Line:
    :param End_Offset:
    :param Text_Pointer: Type: ``File_Buffer_Ptr``
    :param Text_Length:  Type: ``Source_Ptr``
    :return:             Return True in case of success, False in case of failure (the gap is too small).
    """
    return False  # pragma: no cover


@export
def Replace_Text(
    File: SourceFileEntry,
    Start_Line: int,
    Start_Offset: int,
    End_Line: int,
    End_Offset: int,
    Text: str,
) -> bool:
    """Replace [START; END) by TEXT.

    :param File:         File where to replace a text section.
    :param Start_Line:   undocumented
    :param Start_Offset: undocumented
    :param End_Line:     undocumented
    :param End_Offset:   undocumented
    :param Text:         undocumented
    :return:             Return True in case of success, False in case of failure (the gap is too small).
    """
    buffer = Text.encode(ENCODING)
    return _Replace_Text(
        File,
        Start_Line,
        Start_Offset,
        End_Line,
        End_Offset,
        c_char_p(buffer),
        len(buffer),
    )


@export
# @BindToLibGHDL("files_map__editor__fill_text_ptr")
def Fill_Text(File: SourceFileEntry, Text_Pointer, Text_Length: int) -> None:
    """Replace the content of :obj:`File` with TEXT.

    .. todo:: Replace ``Text_Pointer`` and ``Text_Length`` with Python string

    :param File:         File where to replace the content.
    :param Text_Pointer: Type: ``File_Buffer_Ptr``
    :param Text_Length:  Type: ``Source_Ptr``
    """
    libghdl.files_map__editor__fill_text_ptr(File, Text_Pointer, Text_Length)


@export
# @BindToLibGHDL("files_map__editor__check_buffer_content")
def Check_Buffer_Content(File: SourceFileEntry, String_Pointer: c_char_p, String_Length: c_uint32) -> None:
    """
    Check that content of :obj:`File` is STR[1 .. STR_LEN].

    .. todo:: Replace ``String_Pointer`` and ``String_Length`` with Python string

    :param File:           File to check the content.
    :param String_Pointer: Type: ``File_Buffer_Ptr``
    :param String_Length:  Type: ``Source_Ptr``
    """
    libghdl.files_map__editor__check_buffer_content(File, String_Pointer, String_Length)


@export
@BindToLibGHDL("files_map__editor__copy_source_file")
def Copy_Source_File(Dest: SourceFileEntry, Src: SourceFileEntry) -> None:
    """
    Copy content of :obj:`Src` to :obj:`Dest`.

    .. warning:: The size of :obj:`Dest` must be large enough.

    Clear lines table of :obj:`Dest`.
    """



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Package module:   DOM: Interface items (e.g. generic or port)
#
# License:
# ============================================================================
#  Copyright (C) 2019-2022 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================
from typing import List

from pyTooling.Decorators import export

from pyVHDLModel.Symbol import Symbol
from pyVHDLModel.Interface import GenericInterfaceItemMixin, ParameterInterfaceItemMixin
from pyVHDLModel.Subprogram import Procedure as VHDLModel_Procedure, Function as VHDLModel_Function

from pyGHDL.libghdl._types import Iir
from pyGHDL.libghdl.vhdl import nodes
from pyGHDL.dom import DOMMixin
from pyGHDL.dom._Utils import GetNameOfNode, GetDocumentationOfNode
from pyGHDL.dom.Symbol import SimpleSubtypeSymbol


@export
class Function(VHDLModel_Function, DOMMixin):
    def __init__(
        self,
        node: Iir,
        functionName: str,
        returnType: Symbol,
        genericItems: List[GenericInterfaceItemMixin] = None,
        parameterItems: List[ParameterInterfaceItemMixin] = None,
        documentation: str = None,
    ) -> None:
        super().__init__(functionName, documentation)
        DOMMixin.__init__(self, node)

        # TODO: move to model
        self._genericItems = [] if genericItems is None else [g for g in genericItems]
        self._parameterItems = [] if parameterItems is None else [p for p in parameterItems]
        self._returnType = returnType

    @classmethod
    def parse(cls, functionNode: Iir) -> "Function":
        from pyGHDL.dom._Translate import (
            GetName,
            GetGenericsFromChainedNodes,
            GetParameterFromChainedNodes,
        )

        functionName = GetNameOfNode(functionNode)
        documentation = GetDocumentationOfNode(functionNode)

        generics = GetGenericsFromChainedNodes(nodes.Get_Generic_Chain(functionNode))
        parameters = GetParameterFromChainedNodes(nodes.Get_Interface_Declaration_Chain(functionNode))

        returnType = nodes.Get_Return_Type_Mark(functionNode)
        returnTypeName = GetName(returnType)
        returnTypeSymbol = SimpleSubtypeSymbol(returnType, returnTypeName)

        return cls(functionNode, functionName, returnTypeSymbol, generics, parameters, documentation)


@export
class Procedure(VHDLModel_Procedure, DOMMixin):
    def __init__(
        self,
        node: Iir,
        procedureName: str,
        genericItems: List[GenericInterfaceItemMixin] = None,
        parameterItems: List[ParameterInterfaceItemMixin] = None,
        documentation: str = None,
    ) -> None:
        super().__init__(procedureName, documentation)
        DOMMixin.__init__(self, node)

        # TODO: move to model
        self._genericItems = [] if genericItems is None else [g for g in genericItems]
        self._parameterItems = [] if parameterItems is None else [p for p in parameterItems]

    @classmethod
    def parse(cls, procedureNode: Iir) -> "Procedure":
        from pyGHDL.dom._Translate import (
            GetGenericsFromChainedNodes,
            GetParameterFromChainedNodes,
        )

        procedureName = GetNameOfNode(procedureNode)
        documentation = GetDocumentationOfNode(procedureNode)

        generics = GetGenericsFromChainedNodes(nodes.Get_Generic_Chain(procedureNode))
        parameters = GetParameterFromChainedNodes(nodes.Get_Interface_Declaration_Chain(procedureNode))

        return cls(procedureNode, procedureName, generics, parameters, documentation)



# =============================================================================
#               ____ _   _ ____  _       _ _ _           _         _ _
#  _ __  _   _ / ___| | | |  _ \| |     | (_) |__   __ _| |__   __| | |
# | '_ \| | | | |  _| |_| | | | | |     | | | '_ \ / _` | '_ \ / _` | |
# | |_) | |_| | |_| |  _  | |_| | |___ _| | | |_) | (_| | | | | (_| | |
# | .__/ \__, |\____|_| |_|____/|_____(_)_|_|_.__/ \__, |_| |_|\__,_|_|
# |_|    |___/                                     |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Package module:   Python binding and low-level API for shared library 'libghdl'.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================
#
from ctypes import (
    c_int32,
    c_uint32,
    c_char_p,
    c_void_p,
    c_bool,
    c_double,
    Structure,
    c_char,
    c_uint64,
    c_int64,
)
from enum import IntEnum
from functools import wraps
from typing import Callable, List, Dict, Any, TypeVar

from pyTooling.Decorators import export

from pyGHDL.libghdl import libghdl, LibGHDLException


@export
def EnumLookupTable(cls) -> Callable:
    """
    Decorator to precalculate an enum lookup table (LUT) for enum position to
    enum literal name.

    :param cls: Enumerator class for which a LUT shall be pre-calculated.
    """

    def decorator(func) -> Callable:
        def gen() -> List[str]:
            d = [e for e in dir(cls) if e[0].isupper() and e[0] != "_"]
            res = [None] * len(d)
            for e in d:
                res[getattr(cls, e)] = e
            return res

        __lut = gen()

        @wraps(func)
        def wrapper(id: int) -> str:
            # function that replaces the placeholder function
            return __lut[id]

        return wrapper

    return decorator


def BindToLibGHDL(subprogramName):
    """
    This decorator creates a Python function to interface with subprograms in
    libghdl via :mod:`ctypes`.

    :param subprogramName: Name of the subprogram in *libghdl*.
    """

    def PythonTypeToCtype(typ):
        if typ is None:
            return None
        elif typ is int:
            return c_int32
        elif typ is float:
            return c_double
        elif typ is bool:
            return c_bool
        elif typ is bytes:
            return c_char_p
        elif typ in (c_char, c_char_p, c_uint32, c_void_p):
            return typ
        elif isinstance(typ, TypeVar):
            # Humm, recurse ?
            if typ.__bound__ is int:
                return c_int32
            if typ.__bound__ is float:
                return c_double
            if typ.__bound__ in (
                c_bool,
                c_uint32,
                c_int32,
                c_uint64,
                c_int64,
                c_double,
            ):
                return typ.__bound__
            raise TypeError(f"Unsupported typevar bound to {typ.__bound__!s}")
        elif issubclass(typ, IntEnum):
            return c_int32
        elif issubclass(typ, Structure):
            return typ
        raise TypeError

    def wrapper(func: Callable):
        typeHints: Dict[str, Any] = func.__annotations__
        typeHintCount = len(typeHints)

        if typeHintCount == 0:
            raise ValueError(f"Function {func.__name__} is not annotated with types.")

        try:
            returnType = typeHints["return"]
        except KeyError:
            raise ValueError(f"Function {func.__name__} is not annotated with a return type.")

        if (typeHintCount - 1) != func.__code__.co_argcount:
            raise ValueError(
                f"Number of type annotations ({typeHintCount - 1}) for function '{func.__name__}' does not match number of parameters ({func.__code__.co_argcount})."
            )

        # 		print(typeHints)

        parameters = typeHints.copy()
        del parameters["return"]

        parameterTypes = []
        for parameter in parameters.values():
            try:
                parameterTypes.append(PythonTypeToCtype(parameter))
            except TypeError:
                raise TypeError(f"Unsupported parameter type '{parameter!s}' in function '{func.__name__}'.")

        try:
            resultType = PythonTypeToCtype(returnType)
        except TypeError:
            raise TypeError(f"Unsupported return type '{returnType!s}' in function '{func.__name__}'.")

        functionPointer = getattr(libghdl, subprogramName)
        functionPointer.argtypes = parameterTypes
        functionPointer.restype = resultType

        if isinstance(returnType, type) and issubclass(returnType, IntEnum):

            @wraps(func)
            def inner(*args):
                try:
                    returnValue = functionPointer(*args)
                except OSError as ex:
                    errors = [str(ex)]
                    raise LibGHDLException(
                        f"Caught exception when calling '{subprogramName}' in libghdl.",
                        errors,
                    ) from ex

                return returnType(returnValue)

            return inner
        else:

            @wraps(func)
            def inner(*args):
                try:
                    return functionPointer(*args)
                except OSError as ex:
                    errors = [str(ex)]
                    raise LibGHDLException(
                        f"Caught exception when calling '{subprogramName}' in libghdl.",
                        errors,
                    ) from ex

            return inner

    return wrapper



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Package module:   DOM: VHDL design units (e.g. context or package).
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================

"""
This module contains all DOM classes for VHDL's design units (:class:`context <Entity>`,
:class:`architecture <Architecture>`, :class:`package <Package>`,
:class:`package body <PackageBody>`, :class:`context <Context>` and
:class:`configuration <Configuration>`.


"""
from pyTooling.Decorators import export

from pyVHDLModel.PSLModel import VerificationUnit as VHDLModel_VerificationUnit
from pyVHDLModel.PSLModel import VerificationProperty as VHDLModel_VerificationProperty
from pyVHDLModel.PSLModel import VerificationMode as VHDLModel_VerificationMode
from pyVHDLModel.PSLModel import DefaultClock as VHDLModel_DefaultClock

from pyGHDL.libghdl._types import Iir
from pyGHDL.dom import DOMMixin
from pyGHDL.dom._Utils import GetNameOfNode


@export
class VerificationUnit(VHDLModel_VerificationUnit, DOMMixin):
    def __init__(
        self,
        node: Iir,
        identifier: str,
    ) -> None:
        super().__init__(identifier)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, vunitNode: Iir):
        name = GetNameOfNode(vunitNode)

        # FIXME: needs an implementation

        return cls(vunitNode, name)


@export
class VerificationProperty(VHDLModel_VerificationProperty, DOMMixin):
    def __init__(
        self,
        node: Iir,
        identifier: str,
    ) -> None:
        super().__init__(identifier)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, vpropNode: Iir):
        name = GetNameOfNode(vpropNode)

        # FIXME: needs an implementation

        return cls(vpropNode, name)


@export
class VerificationMode(VHDLModel_VerificationMode, DOMMixin):
    def __init__(
        self,
        node: Iir,
        identifier: str,
    ) -> None:
        super().__init__(identifier)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, vmodeNode: Iir):
        name = GetNameOfNode(vmodeNode)

        # FIXME: needs an implementation

        return cls(vmodeNode, name)


@export
class DefaultClock(VHDLModel_DefaultClock, DOMMixin):
    def __init__(
        self,
        node: Iir,
        identifier: str,
    ) -> None:
        super().__init__(identifier)
        DOMMixin.__init__(self, node)

    @classmethod
    def parse(cls, defaultClockNode: Iir):
        name = GetNameOfNode(defaultClockNode)

        # FIXME: needs an implementation

        return cls(defaultClockNode, name)



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Package module:   DOM: IIR helper functions
#
# License:
# ============================================================================
#  Copyright (C) 2019-2022 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================
from pyTooling.Decorators import export

from pyVHDLModel.Base import Mode

from pyGHDL.libghdl import LibGHDLException, name_table, errorout_memory, files_map, file_comments
from pyGHDL.libghdl._types import Iir
from pyGHDL.libghdl.vhdl import nodes, utils
from pyGHDL.libghdl.vhdl.nodes import Null_Iir
from pyGHDL.dom import DOMException

__MODE_TRANSLATION = {
    nodes.Iir_Mode.In_Mode: Mode.In,
    nodes.Iir_Mode.Out_Mode: Mode.Out,
    nodes.Iir_Mode.Inout_Mode: Mode.InOut,
    nodes.Iir_Mode.Buffer_Mode: Mode.Buffer,
    nodes.Iir_Mode.Linkage_Mode: Mode.Linkage,
}
"""
Translation table if IIR modes to pyVHDLModel mode enumeration values.
"""


@export
def CheckForErrors() -> None:
    """
    Check if an error occurred in libghdl and raise an exception if so.

    **Behavior:**

    1. read the error buffer and clear it afterwards.
    2. convert it into a list of internal messages for a :exc:`LibGHDLException`.
    3. raise a :exc:`DOMException` with a nested :exc:`LibGHDLException` as a ``__cause__``.

    :raises DOMException: If an error occurred in libghdl.
    """
    errorCount = errorout_memory.Get_Nbr_Messages()
    if errorCount != 0:
        errors = []
        for i in range(errorCount):
            rec = errorout_memory.Get_Error_Record(i + 1)
            # FIXME: needs help from @tgingold
            fileName = "????"  # name_table.Get_Name_Ptr(files_map.Get_File_Name(rec.file))
            message = errorout_memory.Get_Error_Message(i + 1)

            errors.append(f"{fileName}:{rec.line}:{rec.offset}: {message}")

        errorout_memory.Clear_Errors()

        raise DOMException("Error raised in libghdl.") from LibGHDLException("libghdl: Internal error.", errors)


@export
def GetIirKindOfNode(node: Iir) -> nodes.Iir_Kind:
    """Return the kind of a node in the IIR tree.

    :returns:           The IIR kind of a node.
    :raises ValueError: If parameter ``node`` is :data:`~pyGHDL.libghdl.vhdl.nodes.Null_Iir`.
    """
    if node == Null_Iir:
        raise ValueError("GetIirKindOfNode: Parameter 'node' must not be 'Null_Iir'.")

    kind: int = nodes.Get_Kind(node)
    return nodes.Iir_Kind(kind)


@export
def GetNameOfNode(node: Iir) -> str:
    """Return the Python string from node ``node`` identifier.

    :raises ValueError: If parameter ``node`` is :data:`~pyGHDL.libghdl.vhdl.nodes.Null_Iir`.
    """
    if node == Null_Iir:
        raise ValueError("GetNameOfNode: Parameter 'node' must not be 'Null_Iir'.")

    identifier = utils.Get_Source_Identifier(node)
    return name_table.Get_Name_Ptr(identifier)


@export
def GetDocumentationOfNode(node: Iir) -> str:
    file = files_map.Location_To_File(nodes.Get_Location(node))
    idx = file_comments.Find_First_Comment(file, node)
    documentation = []
    while idx != file_comments.No_Comment_Index:
        documentation.append(file_comments.Get_Comment(file, idx))
        idx = file_comments.Get_Next_Comment(file, idx)

    return "\n".join(documentation)


@export
def GetModeOfNode(node: Iir) -> Mode:
    """Return the mode of a ``node``.

    :raises ValueError:   If parameter ``node`` is :data:`~pyGHDL.libghdl.vhdl.nodes.Null_Iir`.
    :raises DOMException: If mode returned by libghdl is not known by :data:`__MODE_TRANSLATION`.
    """
    if node == Null_Iir:
        raise ValueError("GetModeOfNode: Parameter 'node' must not be 'Null_Iir'.")

    try:
        return __MODE_TRANSLATION[nodes.Get_Mode(node)]
    except KeyError as ex:
        raise DOMException(f"Unknown mode '{ex.args[0]}'.") from ex



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#   Unai Martinez-Corral
#
# Testsuite:        Parse files from ieee2008 directory.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2022 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================
from pathlib import Path

from pytest import mark

from pyGHDL.dom.NonStandard import Design, Document


if __name__ == "__main__":
    print("ERROR: you called a testcase declaration file as an executable module.")
    print("Use: 'python -m unitest <testcase module>'")
    exit(1)


_GHDL_ROOT = Path(__file__).parent.parent.parent.parent.resolve()
_LIBRARIES_ROOT = _GHDL_ROOT / "libraries"

_IEEE2008_ROOT = _LIBRARIES_ROOT / "ieee2008"
_MENTOR_ROOT = _LIBRARIES_ROOT / "mentor"
_SYNOPSYS_ROOT = _LIBRARIES_ROOT / "synopsys"
_VITAL_ROOT = _LIBRARIES_ROOT / "vital2000"


design = Design()


@mark.parametrize("file", [str(f.relative_to(_IEEE2008_ROOT)) for f in _IEEE2008_ROOT.glob("*.vhdl")])
def test_IEEE2008(file):
    print()

    filePath = _IEEE2008_ROOT / file

    lib = design.GetLibrary("ieee2008")
    document = Document(filePath)
    design.AddDocument(document, lib)
    print(f"{document.Path}:")
    for warning in document._warnings:
        print(f"  {warning}")


@mark.parametrize("file", [str(f.relative_to(_MENTOR_ROOT)) for f in _MENTOR_ROOT.glob("*.vhdl")])
def test_Mentor(file):
    print()

    filePath = _MENTOR_ROOT / file

    lib = design.GetLibrary("mentor")
    document = Document(filePath)
    design.AddDocument(document, lib)
    print(f"{document.Path}:")
    for warning in document._warnings:
        print(f"  {warning}")


@mark.parametrize("file", [str(f.relative_to(_SYNOPSYS_ROOT)) for f in _SYNOPSYS_ROOT.glob("*.vhdl")])
def test_Synopsys(file):
    print()

    filePath = _SYNOPSYS_ROOT / file

    lib = design.GetLibrary("synopsys")
    document = Document(filePath)
    design.AddDocument(document, lib)
    print(f"{document.Path}:")
    for warning in document._warnings:
        print(f"  {warning}")


@mark.xfail(reason="Needs further investigations.")
@mark.parametrize("file", [str(f.relative_to(_VITAL_ROOT)) for f in _VITAL_ROOT.glob("*.vhdl")])
def test_Vital(file):
    print()

    filePath = _VITAL_ROOT / file

    lib = design.GetLibrary("vital")
    document = Document(filePath)
    design.AddDocument(document, lib)
    print(f"{document.Path}:")
    for warning in document._warnings:
        print(f"  {warning}")
